/*******************************************************************************

        @file TextReader.d
        
        Copyright (c) 2004 Kris Bell
        
        This software is provided 'as-is', without any express or implied
        warranty. In no event will the authors be held liable for damages
        of any kind arising from the use of this software.
        
        Permission is hereby granted to anyone to use this software for any 
        purpose, including commercial applications, and to alter it and/or 
        redistribute it freely, subject to the following restrictions:
        
        1. The origin of this software must not be misrepresented; you must 
           not claim that you wrote the original software. If you use this 
           software in a product, an acknowledgment within documentation of 
           said product would be appreciated but is not required.

        2. Altered source versions must be plainly marked as such, and must 
           not be misrepresented as being the original software.

        3. This notice may not be removed or altered from any distribution
           of the source.

        4. Derivative works are permitted, but they must carry this notice
           in full and credit the original source.


                        ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


        @version        Initial version, March 2004      
        @author         Kris


*******************************************************************************/

module mango.io.TextReader;

private import  mango.io.Reader,
                mango.io.Exception;

private import  mango.text.Token;

private import  mango.convert.Type,
                mango.convert.Atoi,
                mango.convert.Double;

/*******************************************************************************

        Grok readable input from a stream. All input is tokenized from the
        associated buffer, and converted as necessary into the destination 
        location. A SpaceTokenizer is used by default, but you can choose
        an alternative (such as a comma-delimiting tokenizer).

*******************************************************************************/

class TextReaderTemplate(T) : Reader
{
        private DoubleTemplate!(T)      dbl;
        private AtoiTemplate!(T)        atoi;
        private TokenTemplate!(T)       token;

        /***********************************************************************
        
                Construct a TextReader on the provided buffer, using the
                specified Tokenizer instead of the default one.

        ***********************************************************************/

        this (IBuffer buffer, TokenTemplate!(T) token)
        {
                super (buffer);
                this.token = token;
                token.setRefill (&refill);
        }
        
        /***********************************************************************
                
                Construct a TextReader upon the buffer associated with the
                given conduit.

        ***********************************************************************/

        this (IConduit conduit, TokenTemplate!(T) token)
        {
                this (new Buffer(conduit), token);
        }

        /***********************************************************************
        
                Is this Reader text oriented?

        ***********************************************************************/

        bool isTextBased()
        {
                return true;
        }

        /***********************************************************************
        
        ***********************************************************************/

        override uint read (void* src, uint bytes, uint type)
        {
                int length = bytes;

                // get width of elements (note: does not work for bit[])
                int width = Type.widths[type];

                // for all bytes in source ...
                while (bytes)
                      {
                      T[] t = next.get ();

                      switch (type)
                             {
                             case Type.Bool:
                                  *cast(bool*) src = cast(bool) (t == "true");
                                  break;

                             case Type.Byte:
                             case Type.UByte:
                                  *cast(ubyte*) src = cast(ubyte) atoi.parse (t);
                                  break;

                             case Type.Short:
                             case Type.UShort:
                                  *cast(ushort*) src = cast(ushort) atoi.parse (t);
                                  break;

                             case Type.Int:
                             case Type.UInt:
                                  *cast(uint*) src = cast(uint) atoi.parse (t);
                                  break;

                             case Type.Long:
                             case Type.ULong:
                                  *cast(long*) src = atoi.parse (t);
                                  break;

                             case Type.Float:
                                  *cast(float*) src = dbl.parse (t);
                                  break;

                             case Type.Double:
                                  *cast(double*) src = dbl.parse (t);
                                  break;

                             case Type.Real:
                                  *cast(real*) src = dbl.parse (t);
                                  break;

                             default:
                                  error ("unknown type handed to TextReader.read()");
                             }

                      // bump counters and loop around for next instance
                      bytes -= width;
                      src += width;
                      }

                return length;
        }

        /***********************************************************************
        
                Intercept single char decoding

        ***********************************************************************/

        protected override IReader decode (void* dst, uint bytes, uint type) 
        { 
                next();
                buffer.skip (token.index - buffer.getPosition);
                return super.decode (dst, bytes, type);
        }

        /***********************************************************************
        
                Intercept char-array decoding

        ***********************************************************************/

        protected override uint count (uint elements)
        {
                nextToken();
                buffer.skip (token.index - buffer.getPosition);
                return token.get.length;
        }

        /***********************************************************************
        
                Internal method to capture the next token.

        ***********************************************************************/

        private final TokenTemplate!(T) nextToken ()
        {
                if (! token.next)
                      error ("unexpected end of input"); 
                token.trim ();
                return token;
        }

        /***********************************************************************
        
                Internal method to isolate the next token, and check its
                length. This is used when converting to non-char types, 
                such as integers ~ an empty token is illegal, since we
                don't support default values.

                @todo: should the empty-token rule be relaxed?

        ***********************************************************************/

        private final TokenTemplate!(T) next ()
        {
                if (nextToken.get.length == 0)
                    error ("attempt to parse an invalid empty token");
                return token;    
        }
        
        /***********************************************************************
        
                Throw an exception if the input requested is not available.
                Reader instances expect the input to be available, so you 
                might think of them as being applied to a fixed-format 
                file. Tokenizers, on the other hand, are more flexible in
                that they return true until no more tokens are available.
                      
                Readers take care of assignment to user-space variables, 
                whereas Tokenizers just make the raw content available.
                
        ***********************************************************************/

        private static final void error (char[] msg)
        {
                throw new TokenException (msg);
        }

        /***********************************************************************
 
                Refill the token content from our buffer. Returns false 
                upon reaching EOF

        ***********************************************************************/

        private bool refill (TokenTemplate!(T) token)
        {
                bool more;

                buffer.skip (token.eaten - buffer.getPosition);
                if (buffer.getConduit)
                    more = cast(bool) (buffer.fill() != IConduit.Eof);
                else
                   more = cast(bool) (buffer.readable > 0);
                token.prime (buffer.toString);
                return  more;
        }
}


// convenience alias
alias TextReaderTemplate!(char) TextReader;